@@ -1,11 +1,52 @@
 """
-`Compiled` is a trait indicating that any `:call` expressions should be evaluated
-using Julia's normal compiled-code evaluation. The alternative is to pass `stack=Frame[]`,
-which will cause all calls to be evaluated via the interpreter.
+    abstract type Interpreter end
+
+An interpreter that subtypes this type can implement its own evaluation strategies, by
+overloading the certain methods in JuliaInterpreter that are defined for this base type.
+The default behavior of `Interpreter` is same as that of [`RecursiveInterpreter`](@ref),
+meaning it will recursively interpret all `:call` expressions.
 """
-struct Compiled end
+abstract type Interpreter end
+
+"""
+    RecursiveInterpreter <: Interpreter
+
+`RecursiveInterpreter` is an [`Interpreter`](@ref) that recursively interprets any `:call`
+expressions in the code being interpreted.
+
+With this interpreter, code runs in fully interpreted mode; it will never be compiled for execution.
+"""
+struct RecursiveInterpreter <: Interpreter end
+
+"""
+    NonRecursiveInterpreter <: Interpreter
+
+`NonRecursiveInterpreter` is an [`Interpreter`](@ref) that evaluates any `:call` expressions
+in the code being interpreted using Julia's normal code execution engine with the native
+compiler.
+
+`JuliaInterpreter.Compiled` is aliased to `NonRecursiveInterpreter` for backward compatibility.
+"""
+struct NonRecursiveInterpreter <: Interpreter end
+
+"""
+    const Compiled = NonRecursiveInterpreter
+
+As of JuliaInterpreter v0.10, `Compiled` is now an alias for [`NonRecursiveInterpreter`](@ref).
+This remains for backward compatibility for packages using `Compiled`, and may be removed or
+redefined as a completely different type in v0.11 or later.
+"""
+const Compiled = NonRecursiveInterpreter # for backward compatibility
 Base.similar(::Compiled, sz) = Compiled()  # to support similar(stack, 0)
 
+"""
+    method_table(interpreter::Interpreter) -> mt::Union{Nothing,MethodTable}
+
+Configures the method table used for method lookups performed by the interpreter.
+Uses the global method table by default.
+"""
+method_table(::Interpreter) = nothing
+
 # Our own replacements for Core types. We need to do this to ensure we can tell the difference
 # between "data" (Core types) and "code" (our types) if we step into Core.Compiler
 struct SSAValue
@@ -123,6 +164,13 @@ function is_breakpoint_expr(ex::Expr)
     q = ex.args[3]
     return isa(q, QuoteNode) && q.value === :__BREAKPOINT_MARKER__
 end
+
+@static if isbindingresolved_deprecated
+    is_breakpoint_marker(stmt) = is_global_ref(stmt, JuliaInterpreter, :__BREAK_POINT_MARKER__)
+else
+    is_breakpoint_marker(stmt) = stmt === __BREAK_POINT_MARKER__
+end
+
 function FrameCode(scope, src::CodeInfo; generator=false, optimize=true)
     if optimize
         src, methodtables = optimize!(copy(src), scope)
@@ -132,7 +180,7 @@ function FrameCode(scope, src::CodeInfo; generator=false, optimize=true)
     end
     breakpoints = Vector{BreakpointState}(undef, length(src.code))
     for (i, pc_expr) in enumerate(src.code)
-        if lookup_stmt(src.code, pc_expr) === __BREAK_POINT_MARKER__
+        if is_breakpoint_marker(lookup_stmt(src.code, pc_expr))
             breakpoints[i] = BreakpointState()
             src.code[i] = nothing
         end
@@ -257,7 +305,7 @@ mutable struct Frame
     world::UInt
 end
 function Frame(framecode::FrameCode, framedata::FrameData, pc=1, caller=nothing,
-               world=@static isdefined(Base, :tls_world_age) ? Base.tls_world_age() : Base.get_world_counter())
+               world=@static isdefinedglobal(Base, :tls_world_age) ? Base.tls_world_age() : Base.get_world_counter())
     if length(junk_frames) > 0
         frame = pop!(junk_frames)
         frame.framecode = framecode
